#ifndef S_XSLT_INIT_FUNCTION
#define S_XSLT_INIT_FUNCTION

#include <Include/PlatformDefinitions.hpp>
#include <util/PlatformUtils.hpp>
#include <XalanTransformer/XalanTransformer.hpp>

#include "RSCommon.h"
#undef error

#include "SXSLTUtils.hpp"

class RInitFunction : public Function {

   /* This is the version that gets called with arguments.
      We copy each of these arguments into an array of char * if the arguments are strings.
      Otherwise, we discard them.
    */
  virtual XObjectPtr
  execute(
      XPathExecutionContext& executionContext,
      XalanNode* context,
      int  /* opPos */,
      const XObjectArgVectorType&    args) {

    int numArgs =  args.size() + 1;
    char **sArgs = (char **) malloc(numArgs * sizeof(char *));

     numArgs = 0;
     sArgs[numArgs++] = strdup("SXalan");

     for(int i = 0; i < args.size(); i++) {
      if(args[i]->getType() == XObject::eTypeString)
        sArgs[numArgs++] = getString(args[i]->str());
      else {
       cerr << "Skipping argument " << i << endl;
      }
     }
    
     init_R(numArgs, sArgs);
     
     return(executionContext.getXObjectFactory().createNumber(1)); 
  }

   /*
        This is the version of the function that is called with no arguments and so it 
        just calls init_R() to use the default arguments.
    */
  virtual XObjectPtr
  execute(XPathExecutionContext&			executionContext,
  	  XalanNode*						context) {

     init_R();
     return(executionContext.getXObjectFactory().createNumber(1));
  }

#if defined(XALAN_NO_COVARIANT_RETURN_TYPE)
  virtual Function*
#else
  virtual FunctionRCall*
#endif
  clone() const
  {
    return new RInitFunction(*this);
  }
};

#endif
